"""Tests for quorum functions"""
import pytest
from .utils.sets import deepfreezesets
from .quorum_slice_definition import generate_quorum_slices, get_normalized_definition, \
    remove_from_definition, satisfies_definition

DEFINITION = {'threshold': 2, 'nodes': {'A', 'B', 'C'}, 'children_definitions': []}
DEFINITION_WITHOUT_B = {'threshold': 1, 'nodes': {'A', 'C'}, 'children_definitions': []}

@pytest.mark.parametrize('definition,node,expected', [
    (DEFINITION, 'B', DEFINITION_WITHOUT_B),
    (
        {'threshold': 2, 'nodes': {'D', 'E'}, 'children_definitions': [DEFINITION]},
        'B',
        {'threshold': 2, 'nodes': {'D', 'E'}, 'children_definitions': [DEFINITION_WITHOUT_B]}
    )
])
def test_removal(definition, node, expected):
    """Test remove_from_definition()"""
    result = remove_from_definition(definition, node)
    assert result == expected


def test_normalization():
    """Test get_normalized_definition()"""
    normalized_definition = get_normalized_definition(DEFINITION, 'B')
    expected_definition = {
        'threshold': 2,
        'nodes': {'B'},
        'children_definitions': [DEFINITION_WITHOUT_B]
    }
    assert normalized_definition == expected_definition

def test_qslice_generation():
    """Test generate_quorum_slices()"""
    expected_sets_economic = deepfreezesets([{'A', 'B'}, {'A', 'C'}, {'B', 'C'}])
    result_economic = generate_quorum_slices(DEFINITION)
    assert deepfreezesets(result_economic) == expected_sets_economic
    expected_sets_full = set(expected_sets_economic)
    expected_sets_full.add(frozenset(['A', 'B', 'C']))
    result_full = generate_quorum_slices(DEFINITION, mode='full')
    assert deepfreezesets(result_full) == expected_sets_full

def test_satisfies_definition():
    """Test satisfies_definition()"""
    assert satisfies_definition({'A', 'C'}, DEFINITION) is True
    assert satisfies_definition({'A'}, DEFINITION) is False
    nested_definition = {'threshold': 2, 'nodes': {'D'}, 'children_definitions': [DEFINITION]}
    assert satisfies_definition({'A', 'C', 'D'}, nested_definition) is True
    assert satisfies_definition({'A', 'C'}, nested_definition) is False



"""Test sets utilities"""
from .sets import deepfreezesets, powerset

def test_powerset_list():
    """Test powerset() with a list"""
    result = powerset(['a', 'b'])
    expected = deepfreezesets([{}, {'a'}, {'b'}, {'a', "b"}])
    assert frozenset(result) == expected

def test_powerset_set():
    """Test powerset() with a set"""
    result = powerset({'a', 'b'})
    expected = deepfreezesets([{}, {'a'}, {'b'}, {'a', "b"}])
    assert frozenset(result) == expected



"""Test for Torstens's quorum enumeration"""
from .quorums import enumerate_quorums, contains_slice


def test_enumerate_quorums():
    """Test enumerate_quorums() with simple example"""
    slices_by_node = {
        1: [{1, 2, 3, 7}],
        2: [{1, 2, 3, 7}],
        3: [{1, 2, 3, 7}],
        4: [{4, 5, 6, 7}],
        5: [{4, 5, 6, 7}],
        6: [{4, 5, 6, 7}],
        7: [{7}],
    }

    def ex28_fbas(nodes_subset, node) -> bool:
        return contains_slice(nodes_subset, slices_by_node, node)

    quorums = list(enumerate_quorums((ex28_fbas, {1, 2, 3, 4, 5, 6, 7})))
    assert set(quorums) == {frozenset({7}),
                            frozenset({4, 5, 6, 7}),
                            frozenset({1, 2, 3, 7}),
                            frozenset({1, 2, 3, 4, 5, 6, 7})
                            }


def test_enumerate_quorums_stellar_core():
    """Test enumerate_quorums() with stellar core style fbas"""
    # init test:
    stellar_core_orgs = [
        {'name': "B", 'nodes': ["1", "2", "3"], 'limit': 2},
        {'name': "A", 'nodes': ["1", "2", "3"], 'limit': 2},
        {'name': "C", 'nodes': ["1", "2", "3"], 'limit': 2},
        {'name': "D", 'nodes': ["1", "2", "3"], 'limit': 2},
        {'name': "E", 'nodes': ["1", "2", "3"], 'limit': 2},
        {'name': "F", 'nodes': ["1", "2", "3", "4", "5"], 'limit': 3}
    ]
    stellar_core_nodes = set()
    for org in stellar_core_orgs:
        name, nodes = org['name'], org['nodes']
        for node in nodes:
            stellar_core_nodes.add(name + node)

    threshold = 5

    def stellar_core(subset: set, _: str) -> bool:
        sufficient_orgs = 0
        for org in stellar_core_orgs:
            name, nodes, limit = org['name'], org['nodes'], org['limit']
            sufficient_nodes = 0
            for org_node in nodes:
                node = str(name + org_node)
                if node in subset:
                    sufficient_nodes += 1
            if sufficient_nodes >= limit:
                sufficient_orgs += 1
        return sufficient_orgs >= threshold

    quorums = list(enumerate_quorums((stellar_core, stellar_core_nodes)))
    assert len(quorums) == 114688



"""Methods for quorum slices"""

from itertools import chain

def get_dependencies_by_node(quorum_slices_by_node):
    """Get the dependency graph from quorum slices"""
    return {
        node: frozenset(chain(*quorum_slices)).difference([node])
        for node, quorum_slices in quorum_slices_by_node.items()
    }



"""Utilities for sets"""
from itertools import chain, combinations

def deepfreezesets(sets_iterable):
    """Deep-freeze a list of sets"""
    return frozenset([frozenset(list(el)) for el in sets_iterable])

def powerset(iterable):
    """Return the power set of the input iterable"""
    frozen_iterable_set = frozenset(iterable)
    all_combinations = [combinations(frozen_iterable_set, size)
                        for size in range(len(frozen_iterable_set) + 1)]
    return [frozenset(combination) for combination in chain(*all_combinations)]



"""Tests for quorum functions"""
from .utils.sets import deepfreezesets
from .quorum import get_minimal_quorum_intersection, quorum_intersection

QSET_DEFINITION = {'threshold': 2, 'nodes': {'A', 'B', 'C'}, 'children_definitions': []}
QSET_DEFINITION_WITHOUT_B = {'threshold': 1, 'nodes': {'A', 'C'}, 'children_definitions': []}

def test_quorum_intersection():
    """Test quorum_intersection()"""
    quorums = deepfreezesets([{'A', 'B'}, {'A', 'C'}, {'B', 'C'}])
    has_intersection, intersection_quorums, split_quorums = \
        quorum_intersection(quorums)
    assert has_intersection is True
    assert len(intersection_quorums) == 3
    assert not split_quorums

def test_quorum_intersection_fail():
    """Test quorum_intersection()"""
    quorums = deepfreezesets([{'A', 'B'}, {'B', 'C'}, {'C', 'D'}])
    has_intersection, intersection_quorums, split_quorums = \
        quorum_intersection(quorums)
    assert has_intersection is False
    assert len(intersection_quorums) == 2
    assert len(split_quorums) == 1
    assert frozenset(split_quorums[0]) == deepfreezesets([{'A', 'B'}, {'C', 'D'}])

def test_minimal_intersection():
    """Test get_minimal_quorum_intersection()"""
    quorums = deepfreezesets([{'A', 'B'}, {'A', 'C'}, {'B', 'C'}])
    intersection, quorum_a, quorum_b = get_minimal_quorum_intersection(quorums)
    assert len(intersection) == 1
    assert len(quorum_a.intersection(quorum_b)) == 1

def test_minimal_intersection_fail():
    """Test get_minimal_quorum_intersection()"""
    quorums = deepfreezesets([{'A', 'B'}, {'B', 'C'}, {'C', 'D'}])
    intersection, quorum_a, quorum_b = get_minimal_quorum_intersection(quorums)
    assert intersection is None
    assert quorum_a.intersection(quorum_b) == set()



"""Tests for Torstens's intactNode Algorithm"""
from stellarobservatory.intactness import intact_nodes
from stellarobservatory.quorums import contains_slice


def test_intact_nodes():
    """Test intact_nodes (Example 4.13)"""
    slices_by_node = {
        'A': [{'A', 'B'}],
        'B': [{'A', 'B', 'C', 'D'}],
        'C': [{'B', 'C'}, {'C', 'D'}],
        'D': [{'B', 'D'}, {'C', 'D'}]
    }

    def is_slice_contained(nodes_subset, node) -> bool:
        return contains_slice(nodes_subset, slices_by_node, node)

    res = intact_nodes((is_slice_contained, {'A', 'B', 'C', 'D'}), {'A'})
    assert res == {'C', 'D'}



"""Torstens's quorum intersection checker (a Lachowski variant)"""
from typing import Callable, Set, Type, Tuple

from stellarobservatory.quorums import greatest_quorum


def quorum_intersection(fbas: Tuple[Callable[[Set[Type], Type], bool], Set[Type]]):
    """Takes an FBAS with set of nodes V and returns True iff F has quorum intersection.
    It prints two disjoint quorums otherwise."""
    is_slice_contained, all_nodes = fbas
    len_all_nodes = len(all_nodes)
    for quorum in traverse_min_quorums(is_slice_contained, set(), all_nodes, len_all_nodes):
        greatest_q = greatest_quorum(is_slice_contained,
                                     all_nodes.difference(quorum), set())
        if greatest_q != set():
            return False, quorum, greatest_q
    return True


def contains_proper_sub_quorum(is_slice_contained: Callable[[Set[Type], Type], bool],
                               subset_nodes: set):
    """Takes an FBAS with set of nodes V; and a subset U of V and
    returns whether there is a quorum Q not fully contained U"""
    for node in subset_nodes:
        if greatest_quorum(is_slice_contained,
                           subset_nodes.difference({node}), set()) != set():
            return True
    return False


def traverse_min_quorums(is_slice_contained: Callable[[Set[Type], Type], bool],
                         committed: set,  # U
                         remaining: set,  # R
                         len_all_nodes: int):  # |V|
    """Enumerate all min quorums Q with U ⊆ Q ⊆ U∪R and |Q|≤|V|/2"""
    if len(committed) > len_all_nodes / 2:  # if |U|>|V|/2 return
        return
    greatest_q = greatest_quorum(is_slice_contained, committed, set())
    if greatest_q != set():
        if committed == greatest_q and not contains_proper_sub_quorum(is_slice_contained,
                                                                      committed):
            yield committed
    else:
        perimeter = committed.union(remaining)
        if remaining != set() and committed.issubset(greatest_quorum(is_slice_contained,
                                                                     perimeter,
                                                                     remaining)):
            # v ← pick from R:
            # (note pylint complains:
            # Do not raise StopIteration in generator, use return statement instead
            # but this can't happen as current != set())
            # pylint: disable=R1708
            node = next(iter(remaining))
            remaining_without_v = remaining.difference({node})
            yield from traverse_min_quorums(is_slice_contained,
                                            committed,
                                            remaining_without_v,
                                            len_all_nodes)
            yield from traverse_min_quorums(is_slice_contained,
                                            committed.union({node}),
                                            remaining_without_v,
                                            len_all_nodes)


def is_quorum(is_slice_contained: Callable[[Set[Type], Type], bool], nodes_subset: set):
    """
    Check whether nodes_subset is a quorum in FBAS F (implicitly is_slice_contained method).
    """
    return all([
        is_slice_contained(nodes_subset, v)
        for v in nodes_subset
    ])



"""Test graph utilities"""
from .graph import get_transpose_graph, get_indegrees, get_induced_subgraph

GRAPH = {
        1: {2, 3},
        2: {3},
        3: {2}
    }

def test_get_transpose_graph():
    """Test get_transpose_graph() with a graph"""
    transpose = get_transpose_graph(GRAPH)
    assert transpose == {
        1: set(),
        2: {1, 3},
        3: {1, 2}
    }

def test_get_indegrees():
    """Test get_indegrees() with a graph"""
    assert get_indegrees(GRAPH) == {1: 0, 2: 2, 3: 2}

def test_get_induced_subgraph():
    """Test get_induced_subgraph() with a graph"""
    assert get_induced_subgraph(GRAPH, {1, 2, 3}) == GRAPH
    assert get_induced_subgraph(GRAPH, {1}) == {1: set()}
    assert get_induced_subgraph(GRAPH, {1, 2}) == {1: {2}, 2: set()}
    assert get_induced_subgraph(GRAPH, {2, 3}) == {2: {3}, 3: {2}}
    assert get_induced_subgraph(GRAPH, set()) == dict()



"""Quorum slices tests"""
from .quorum_slices import get_dependencies_by_node

def test_get_dependencies_by_node():
    """Test get_dependencies_by_node()"""
    quorum_slices_by_node = {
        1: [[1, 2], [1, 3]],
        2: [[2, 3]],
        3: [[1, 3]],
        4: [[1, 4]]
    }
    dependencies_by_node = get_dependencies_by_node(quorum_slices_by_node)
    assert dependencies_by_node == {
        1: {2, 3},
        2: {3},
        3: {1},
        4: {1}
    }



"""Quorum slice definitions"""

from itertools import chain, combinations, product

def remove_from_definition(definition, node):
    """Return quorum slice definition with the given node removed and the threshold reduced by 1"""
    threshold = definition['threshold']
    nodes = definition['nodes']
    if node in nodes:
        nodes = nodes.copy()
        nodes.remove(node)
        threshold = max(0, threshold - 1)
    children_definitions = [
        remove_from_definition(children_definition, node)
        for children_definition in definition['children_definitions']
        ]
    return {
        'threshold': threshold,
        'nodes': nodes,
        'children_definitions': children_definitions
    }

def get_normalized_definition(definition, node):
    """Returns the node's quorum slice definition as Stellar Core preprocesses it

    The node is required on root level and removed from rest of quorum slice definition"""
    return {
        'threshold': 2,
        'nodes': {node},
        'children_definitions': [remove_from_definition(definition, node)]
    }

def generate_quorum_slices(definition, mode='economic'):
    """Generate all quorum slices for a quorum slice definition

    'economic' mode only returns quorum slices of size equal to the threshold,
    use 'full' to obtain all quorum slices"""
    threshold = definition['threshold']
    nodes = definition['nodes']
    children_definitions = definition['children_definitions']
    max_size = threshold \
        if mode == 'economic' \
        else len(nodes) + len(children_definitions)

    quorum_slice_pools = [[[node]] for node in nodes] + \
        [generate_quorum_slices(children_definition, mode)
         for children_definition in children_definitions]

    quorum_slice_combinations = list(chain(*[
        combinations(quorum_slice_pools, size)
        for size in range(threshold, max_size + 1)
        ]))

    quorum_slice_products = list(chain(*[
        product(*quorum_slice_combination)
        for quorum_slice_combination in quorum_slice_combinations
        ]))

    return [frozenset(chain(*quorum_slice_product))
            for quorum_slice_product in quorum_slice_products]


def satisfies_definition(candidate, definition):
    '''Checks if the candidate set contains a slice for the provided definition'''
    satisfied = len(definition['nodes'].intersection(candidate))
    for children_definition in definition['children_definitions']:
        if satisfies_definition(candidate, children_definition):
            satisfied += 1
    return satisfied >= definition['threshold']



"""Utilities for graphs"""

def get_transpose_graph(graph):
    """Get the transpose graph"""
    transpose = {node: set() for node in graph.keys()}
    for node, target_nodes in graph.items():
        for target_node in target_nodes:
            transpose[target_node].add(node)
    return transpose

def get_indegrees(graph):
    """Get a dict with indegrees for all nodes"""
    transpose = get_transpose_graph(graph)
    return {node: len(target_nodes) for node, target_nodes in transpose.items()}

def get_induced_subgraph(graph, nodes):
    """Get the nodes-induced subgraph G[S] for a graph G and a subset of nodes S"""
    return {node: graph[node].intersection(nodes) for node in nodes}



"""Quorum slices and quorums"""

from itertools import combinations

def quorum_intersection(quorums):
    """Returns whether the quorums have intersection, which intersect, and which do not"""
    intersecting_quorums = []
    split_quorums = []
    for (quorum_a, quorum_b) in combinations(quorums, 2):
        intersection = quorum_a.intersection(quorum_b)
        if intersection:
            intersecting_quorums.append((quorum_a, quorum_b, intersection))
        else:
            split_quorums.append((quorum_a, quorum_b))
    return len(split_quorums) == 0, intersecting_quorums, split_quorums

def get_minimal_quorum_intersection(quorums):
    """Returns a minimal quorum intersection (or None)"""
    minimal_intersection = None
    for (quorum_a, quorum_b) in combinations(quorums, 2):
        intersection = quorum_a.intersection(quorum_b)
        if not intersection:
            return None, quorum_a, quorum_b
        if minimal_intersection is None or len(intersection) < len(minimal_intersection[0]): #pylint: disable=unsubscriptable-object
            minimal_intersection = intersection, quorum_a, quorum_b
    return minimal_intersection



"""Utilities for strongly connected components"""
import numpy
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import connected_components


def get_strongly_connected_components(graph):
    """
    Get strongly connected components for a directed graph

    The returned list of components is in reverse topological order, i.e.,
    such that the nodes in the first component have no dependencies on
    other components.
    """
    nodes = list(graph.keys())
    node_index_by_node = {node: index for index, node in enumerate(nodes)}
    row_indexes = []
    col_indexes = []
    for node, targets in graph.items():
        row_indexes += [node_index_by_node[node]] * len(targets)
        col_indexes += [node_index_by_node[target] for target in targets]
    data = numpy.ones((len(row_indexes)), dtype=int)
    n_nodes = len(nodes)
    csgraph = csr_matrix((data, (row_indexes, col_indexes)), shape=(n_nodes, n_nodes))
    n_components, labels = connected_components(csgraph, directed=True, connection='strong')
    sccs = [[] for i in range(n_components)]
    for index, label in enumerate(labels):
        sccs[label] += [nodes[index]]
    return [frozenset(scc) for scc in sccs]



"""Fetch and process nodes"""
import requests

def get_nodes_from_stellarbeat():
    """Fetch nodes from stellarbeat.io"""
    return requests.get('https://api.stellarbeat.io/v1/nodes').json()

def get_nodes_by_public_key(nodes):
    """Get nodes by public key as a dictionary"""
    return {node['publicKey']: node for node in nodes}

def convert_public_keys_to_names(nodes_by_public_key, public_keys):
    """Convert a set/list of node public keys to a set of names"""
    return {
        nodes_by_public_key[public_key]['name'] if 'name' in nodes_by_public_key[public_key] \
        else public_key \
        for public_key in public_keys
    }

def get_node_dependencies(nodes_by_public_key, public_key, dependencies=None, transitive=True):
    """Get public keys of all nodes a node depends on (also transitively)"""
    if dependencies is None:
        dependencies = set([public_key])

    def traverse_quorum_set_definition(quorum_set_definition):
        """Traverses a qset definition and adds them to the dependencies"""
        for validator in quorum_set_definition['validators']:
            if validator not in dependencies:
                dependencies.add(validator)
                if transitive:
                    get_node_dependencies(nodes_by_public_key, validator, dependencies)
        for inner_quorum_set_definition in quorum_set_definition['innerQuorumSets']:
            traverse_quorum_set_definition(inner_quorum_set_definition)

    traverse_quorum_set_definition(nodes_by_public_key[public_key]['quorumSet'])
    return dependencies

def get_trust_graph(nodes_by_public_key):
    """
    Map each node's public key to its trust graph (the set of nodes it
    references in its quorum slices). The node itself is excluded.
    """
    graph = {}
    for key in nodes_by_public_key:
        nodes_intrans = get_node_dependencies(nodes_by_public_key, key, \
                                          dependencies=set(), transitive=False)
        graph[key] = set(nodes_intrans)
        graph[key].discard(key)
    return graph



"""Nodes tests"""
from .nodes import get_nodes_from_stellarbeat, get_nodes_by_public_key, \
    get_node_dependencies, get_trust_graph

def test_stellarbeat_nodes():
    """Test get_nodes_from_stellarbeat()"""
    nodes = get_nodes_from_stellarbeat()
    assert isinstance(nodes, list)
    for node in nodes:
        assert isinstance(node, dict)
        assert 'publicKey' in node

nodes_ABCDE = [
    {
        'publicKey': 'A',
        'quorumSet': {'threshold': 2, 'validators': ['A', 'B'], 'innerQuorumSets': []}
    },
    {
        'publicKey': 'B',
        'quorumSet': {
            'threshold': 2,
            'validators': ['B'],
            'innerQuorumSets':[
                {'threshold': 1, 'validators': ['A', 'C'], 'innerQuorumSets': []}
            ]
        }
    },
    {
        'publicKey': 'C',
        'quorumSet': {'threshold': 2, 'validators': ['A', 'C'], 'innerQuorumSets': []}
    },
    {
        'publicKey': 'D',
        'quorumSet': {'threshold': 2, 'validators': ['A', 'B', 'C', 'D'], 'innerQuorumSets': []}
    },
    {
        'publicKey': 'E',
        'quorumSet':{
            'threshold': 2,
            'validators': ['A'],
            'innerQuorumSets':[
                {'threshold': 1, 'validators': ['A', 'D'], 'innerQuorumSets': []}
            ]
        }
    },
]

def test_node_dependencies():
    """Test get_node_dependencies()"""
    nodes_by_public_key = get_nodes_by_public_key(nodes_ABCDE)
    dependencies_a = get_node_dependencies(nodes_by_public_key, 'A')
    assert frozenset(dependencies_a) == frozenset(['A', 'B', 'C'])
    dependencies_a_int = get_node_dependencies(nodes_by_public_key, 'A', transitive=False)
    assert frozenset(dependencies_a_int) == frozenset(['A', 'B'])
    dependencies_d = get_node_dependencies(nodes_by_public_key, 'D')
    assert frozenset(dependencies_d) == frozenset(['A', 'B', 'C', 'D'])
    dependencies_e_int = get_node_dependencies(nodes_by_public_key, 'E', transitive=False)
    assert frozenset(dependencies_e_int) == frozenset(['A', 'D', 'E'])

def test_get_trust_graph():
    """Test get_trust_graph()"""
    nodes_by_public_key = get_nodes_by_public_key(nodes_ABCDE)
    trust_graphs = get_trust_graph(nodes_by_public_key)
    assert isinstance(trust_graphs, dict)
    assert len(trust_graphs) == 5
    assert trust_graphs['A'] == frozenset(['B'])
    assert trust_graphs['B'] == frozenset(['A', 'C'])
    assert trust_graphs['C'] == frozenset(['A'])
    assert trust_graphs['D'] == frozenset(['A', 'B', 'C'])
    assert trust_graphs['E'] == frozenset(['A', 'D'])



"""setup"""
import setuptools

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setuptools.setup(
    name="stellar-observatory",
    version="1.0a10",
    author="André Gaul",
    author_email="andre@gaul.io",
    description="Python package for analyzing the Stellar network",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/andrenarchy/stellar-observatory",
    packages=setuptools.find_packages(),
    python_requires="~=3.5",
    install_requires=[
        "numpy~=1.0",
        "requests~=2.0",
        "scipy~=1.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research"
    ],
)



"""Test scc utilities"""
from .scc import get_strongly_connected_components

def test_get_strongly_connected_components():
    """Test get_strongly_connected_components() with a graph"""
    graph = {
        # SCC 1
        1: {2, 3, 4},
        2: {1, 3, 4},
        3: {1, 2, 4},
        4: {1, 2, 3},
        # SCC 2
        5: {1, 6},
        6: {7},
        7: {8},
        8: {9},
        9: {5}
    }
    sccs = get_strongly_connected_components(graph)
    assert sccs == [{1, 2, 3, 4}, {5, 6, 7, 8, 9}]



"""Torstens's quorum enumeration"""

from typing import Callable, Tuple, Set
from typing import Type


# Allow for defining an FBAS as a function: (set<T>, T, set<T>) -> bool.
# This function returns True, iff the FBAS has a slice for the given node T in the given
# set.
def enumerate_quorums(fbas: Tuple[Callable[[Set[Type], Type], bool], Set[Type]]):
    """Enumerate all quorums of FBAS F (given by the pair (function(set<T>, T) -> bool, set))."""
    (is_slice_contained, all_nodes) = fbas
    return traverse_quorums(is_slice_contained, set(), all_nodes)


def traverse_quorums(is_slice_contained: Callable[[Set[Type], Type], bool],
                     committed: set,
                     remaining: set):
    """Given a FBAS F (by is_slice_contained) with set of nodes V
    and given the sets: committed ⊆ V; R ⊆ V\\committed,
    enumerate all quorums Q of F with committed ⊆ Q ⊆ committed ∪ remaining"""
    perimeter = committed.union(remaining)
    greatest_q = greatest_quorum(is_slice_contained, perimeter, committed)
    if greatest_q == set():
        return
    yield frozenset(greatest_q)

    current = greatest_q.difference(committed)
    while not current == set():
        # v ← pick from W = current
        # (note pylint complains:
        # Do not raise StopIteration in generator, use return statement instead
        # but this can't happen as current != set())
        # pylint: disable=R1708
        node = next(iter(current))
        yield from traverse_quorums(is_slice_contained,
                                    greatest_q.difference(current),
                                    current.difference({node}))
        current = current.difference({node})


def greatest_quorum(is_slice_contained: Callable[[Set[Type], Type], bool],
                    nodes: set,
                    lower_bound: set):
    """
    Return greatest quorum contained in nodes if it is a super set of lower_bound
    or empty set (if there is no such quorum).
    """
    while True:
        next_u = set()
        for node in nodes:
            if is_slice_contained(nodes, node):
                next_u.add(node)
            else:
                if node in lower_bound:
                    return set()
        if len(nodes) == len(next_u):
            return next_u
        nodes = next_u


def contains_slice(nodes_subset: Set, slices_by_node, node):
    """Check if for the given node quorum slices there is a quorum slice
    contained in the subset of nodes.
    Input: FBAS(V,S) implicitly passed in via slices; nodes_subset ⊆ V; node ∈ V, a set D ⊆ V
    Output: whether node has a quorum slice in the FBAS (without D) contained in nodes_subset"""
    return any(quorum_slice.issubset(nodes_subset)
               for quorum_slice in slices_by_node[node])



"""Algorithm for determining B-intact nodes given a set B of nodes."""
from typing import Tuple, Callable, Set, Type

from stellarobservatory.quorum_intersection import quorum_intersection
from stellarobservatory.quorums import greatest_quorum


def intact_nodes(fbas: Tuple[Callable[[Set[Type], Type], bool], Set[Type]],
                 b_nodes: set):
    """
    Takes an FBAS F (having quorum intersection) with set of nodes V and B ⊆ V and
    returns the set of all B-intact nodes.
    """
    is_slice_contained, all_nodes = fbas
    current = all_nodes.difference(b_nodes)
    while True:
        greatest_q = greatest_quorum(is_slice_contained, current, set())

        def cur_is_slice_contained(candidate, node):
            if node in greatest_q:
                raise ValueError('node not defined in FBAS')
            return is_slice_contained(candidate.union(greatest_q), node)

        cur_fbas = (cur_is_slice_contained, all_nodes.difference(greatest_q))
        # determine whether F^{V\Q} has quorum intersection:
        result = quorum_intersection(cur_fbas)
        if result is True:
            return greatest_q
        # else:
        _, quorum1, quorum2 = result
        current_w1 = greatest_quorum(is_slice_contained, greatest_q.difference(quorum1),
                                     set())
        current_w2 = greatest_quorum(is_slice_contained, greatest_q.difference(quorum2),
                                     set())
        if current_w1 == set():
            current = current_w2
        elif current_w2 == set():
            current = current_w1
        else:
            current = current_w1.intersection(current_w2)



"""Stellar Observatory"""
from . import intactness, nodes, quorum_intersection, quorum_slice_definition, \
    quorum_slices, quorum, quorums, utils

__all__ = ['intactness', 'nodes', 'quorum_intersection',
           'quorum_slice_definition', 'quorum_slices', 'quorum',
           'quorums', 'utils']



"""Test for Torstens's quorum intersection checker (Lachowski variant)"""

from stellarobservatory.quorum_intersection import quorum_intersection, is_quorum
from stellarobservatory.quorums import contains_slice


def test_has_quorum_intersection_false():
    """Test quorum_intersection() == False"""
    # Basically two disjoint quorums {A, B} and {D} (two SCCs)
    slices_by_node = {
        'A': [{'A', 'B'}, {'A', 'C'}, {'A', 'B', 'C'}],
        'B': [{'A', 'B'}],
        'C': [{'A', 'B', 'C', 'D'}],
        'D': [{'D'}]
    }

    def is_slice_contained(nodes_subset, node) -> bool:
        return contains_slice(nodes_subset, slices_by_node, node)

    has_intersection, quorum1, quorum2 = quorum_intersection((is_slice_contained,
                                                              {'A', 'B', 'C', 'D'}))
    assert has_intersection is False
    assert is_quorum(is_slice_contained, quorum1) is True
    assert is_quorum(is_slice_contained, quorum2) is True
    assert quorum1.intersection(quorum2) == set()


def test_has_quorum_intersection_false_in_scc():
    """Test has_quorum_intersection() without intersection inside an scc.
    Disjoint SCCs:  {1, 2}, {3, 4}"""
    slices_by_node = {
        1: [{1, 2}, {1, 3}, {1, 4}],
        2: [{2, 1}, {2, 3}, {2, 4}],
        3: [{1, 3}, {2, 3}, {3, 4}],
        4: [{1, 4}, {2, 4}, {3, 4}]
    }

    def is_slice_contained(nodes_subset, node) -> bool:
        return contains_slice(nodes_subset, slices_by_node, node)

    has_intersection, quorum1, quorum2 = quorum_intersection((is_slice_contained,
                                                              {1, 2, 3, 4}))
    assert has_intersection is False
    assert is_quorum(is_slice_contained, quorum1) is True
    assert is_quorum(is_slice_contained, quorum2) is True
    assert quorum1.intersection(quorum2) == set()


def test_has_quorum_intersection_false_two_max_scc():
    """Test has_quorum_intersection() without intersection with two max scc:
    Disjoint SCCs:  {2}, {3}"""
    slices_by_node = {
        1: [{1, 2, 3}],
        2: [{2}],
        3: [{3}]
    }

    def is_slice_contained(nodes_subset, node) -> bool:
        return contains_slice(nodes_subset, slices_by_node, node)

    has_intersection, quorum1, quorum2 = quorum_intersection((is_slice_contained, {1, 2, 3}))
    assert has_intersection is False
    assert is_quorum(is_slice_contained, quorum1) is True
    assert is_quorum(is_slice_contained, quorum2) is True
    assert quorum1.intersection(quorum2) == set()


def test_has_quorum_intersection_true():
    """Test has_quorum_intersection()"""
    # One SCC containing quorum {A, B}
    quorum_slices_by_node = {
        'A': [{'A', 'B'}, {'A', 'C'}, {'A', 'B', 'C'}],
        'B': [{'A', 'B'}],
        'C': [{'A', 'B', 'C'}]
    }

    def is_slice_contained(nodes_subset, node) -> bool:
        return contains_slice(nodes_subset, quorum_slices_by_node, node)

    assert quorum_intersection((is_slice_contained, {'A', 'B', 'C'})) is True


def test_has_quorum_intersection_true_sccs():
    """Test has_quorum_intersection() with intersection with three sccs"""
    slices_by_node = {
        1: [{1, 2}, {1, 3}],
        2: [{2, 1}, {2, 3}],
        3: [{1, 3}, {2, 3}],
        4: [{1, 4}],
        5: [{2, 5}]
    }

    def is_slice_contained(nodes_subset, node) -> bool:
        return contains_slice(nodes_subset, slices_by_node, node)

    assert quorum_intersection((is_slice_contained, {1, 2, 3, 4, 5})) is True
